'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
/* eslint no-underscore-dangle: 0 */
/* eslint no-unneeded-ternary: 0 */

var fs = require('fs');
var path = require('path');
var EventEmitter = require('events');

var ModokEmitter = new EventEmitter();

/* HELPER */

String.prototype.lengthInUtf8 = function lengthInUtf8() {
  var asciiLength = this.match(/[\u0000-\u007f]/g) ? this.match(/[\u0000-\u007f]/g).length : 0;
  var multiByteLength = encodeURI(this.replace(/[\u0000-\u007f]/g)).match(/%/g) ? encodeURI(this.replace(/[\u0000-\u007f]/g, '')).match(/%/g).length : 0;
  return asciiLength + multiByteLength;
};

function uuid() {
  return Math.random().toString(26).slice(2);
}

function getMemoryStats(store) {
  var values = Array.from(store.values());
  var valuesString = JSON.stringify(values);
  return {
    size: valuesString.lengthInUtf8()
  };
}

function newDatabaseFile(filePath, fileName) {
  return new Promise(function (resolve, reject) {
    var filePathChunks = filePath.split('/');

    var breadcrumb = '';
    for (var i = 0; i < filePathChunks.length; i += 1) {
      if (filePathChunks[i]) {
        breadcrumb += '/' + filePathChunks[i];
        if (!fs.existsSync(breadcrumb)) {
          fs.mkdirSync(breadcrumb);
        }
      }
    }

    fs.writeFile(filePath + '/' + fileName + '.json', JSON.stringify({}), 'utf8', function (err) {
      if (err) {
        return reject(err);
      }
      return resolve();
    });
  });
}

function readDatabaseFile(filePath, fileName) {
  return new Promise(function (resolve, reject) {
    fs.readFile(filePath + '/' + fileName + '.json', 'utf8', function (err, data) {
      if (err) {
        return reject(err);
      }
      return resolve(data);
    });
  });
}

function writeDatabaseFile(filePath, fileName, content) {
  return new Promise(function (resolve, reject) {
    if (filePath && fileName && fs.existsSync(filePath + '/' + fileName + '.json')) {
      fs.writeFile(filePath + '/' + fileName + '.json', JSON.stringify(content), 'utf8', function (err) {
        if (err) {
          return reject(err);
        }
        return resolve(JSON.stringify(content));
      });
      resolve('test');
    } else {
      reject({ message: 'No file write' });
    }
  }).then(null, function () {});
}

function readDatabaseStats(store, filePath, fileName) {
  return new Promise(function (resolve, reject) {
    if (filePath && fileName) {
      fs.open(filePath + '/' + fileName + '.json', 'r', function (err, fd) {
        if (err) {
          return reject(err);
        }

        if (fd) {
          fs.fstat(fd, function (_err, stats) {
            if (_err) {
              return reject(_err);
            }
            return resolve(stats);
          });
        }
        return reject('No file descriptor found');
      });
    }

    return resolve(getMemoryStats(store));
  });
}

function readDatabaseStatsSync(store, filePath, fileName) {
  if (!filePath || !fileName) {
    return getMemoryStats(store);
  }

  var fd = fs.openSync(filePath + '/' + fileName + '.json', 'r');
  return fs.fstatSync(fd);
}

function resolveData(object) {
  var id = object._id;
  if (id === undefined || id === null) {
    id = uuid();
  }
  var data = object;
  data._id = id;
  data.created_at = new Date();
  this.store.set(id, data);

  writeDatabaseFile(this.filePath, this.fileName, Array.from(this.store.values()));
  return this.store.get(id);
}

function isObject(item) {
  return !!item && item.constructor === Object;
}

function isArray(item) {
  return !!item && item.constructor === Array;
}

/* Core */

var Brainstack = {};

var Brain = function Brain(name, config) {
  var _this = this;

  if (config && isObject(config)) {
    this.filePath = typeof config.filepath === 'string' ? path.resolve(config.filepath) : null;
    this.fileName = config.filename ? config.filename : name;
    this.writeToFile = fs && isObject(fs) && !this.filePath;

    if (this.filePath && this.fileName) {
      if (!fs.existsSync(this.filePath + '/' + this.fileName + '.json')) {
        newDatabaseFile(this.filePath, this.fileName).then(function () {
          ModokEmitter.emit('ready');
        });
      } else {
        readDatabaseFile(this.filePath, this.fileName).then(function (content) {
          _this.$insertMany(JSON.parse(content)).then(function () {
            ModokEmitter.emit('ready');
          });
        });
      }
    } else {
      ModokEmitter.emit('ready');
    }
  } else {
    ModokEmitter.emit('ready');
  }

  if (!name) {
    throw Error('Name is required');
  }

  if (!this) {
    if (!Brainstack[name]) {
      throw Error('No database with name ' + name + ' found');
    } else {
      return Brainstack[name];
    }
  } else {
    this.store = new Map();
    this.name = name;
    Brainstack[name] = this;
  }

  return this;
};

/**
 * ready
 * Returns a promise.
 */

Brain.prototype.ready = function ready() {
  return new Promise(function (resolve) {
    ModokEmitter.on('ready', function () {
      return resolve();
    });
  });
}();

/**
 * has / $has
 * Returns a boolean asserting whether a value has been
 * associated to the index in the database or not.
 */

// Synchronous
Brain.prototype.has = function has(_id) {
  return this.store.has(_id);
};

// Asynchronous
Brain.prototype.$has = function $has(_id) {
  var _this2 = this;

  return new Promise(function (resolve) {
    resolve(_this2.store.has(_id));
  });
};

/**
 * insert / $insert
 * Inserts a document or documents into a collection.
 * _id: If the document does not specify an _id field,
 * then modokDB will add the _id field and assign a
 * unique id for the document before inserting.
 * If the document contains an _id field, the _id value
 * must be unique within the collection to avoid overwrites.
 */

// Synchronous
Brain.prototype.insert = function insert(object) {
  if (object && Array.isArray(object)) {
    var list = [];
    for (var i = 0; i < object.length; i += 1) {
      list.push(resolveData.call(this, object[i]));
    }
    return list;
  }
  return resolveData.call(this, object);
};

// Asynchronous
Brain.prototype.$insert = function $insert(object) {
  var _this3 = this;

  return new Promise(function (resolve) {
    resolve(_this3.insert(object));
  });
};

/**
 * insertOne / $insertOne
 * Inserts a document into a collection.
 * _id: If the document does not specify an _id field,
 * then modokDB will add the _id field and assign a
 * unique id for the document before inserting.
 * If the document contains an _id field, the _id value
 * must be unique within the collection to avoid overwrites.
 */

// Synchronous
Brain.prototype.insertOne = function insertOne(object) {
  if (isObject(object)) {
    return this.insert(object);
  }

  return null;
};

// Asynchronous
Brain.prototype.$insertOne = function $insertOne(object) {
  var _this4 = this;

  return new Promise(function (resolve) {
    if (isObject(object)) {
      return resolve(_this4.insert(object));
    }
    return resolve(null);
  });
};

/**
 * insertMany / $insertMany
 * Inserts multiple documents into a collection.
 * _id: If the document does not specify an _id field,
 * then modokDB will add the _id field and assign a
 * unique id for the document before inserting.
 * If the document contains an _id field, the _id value
 * must be unique within the collection to avoid overwrites.
 */

// Synchronous
Brain.prototype.insertMany = function insertMany(object) {
  if (isArray(object)) {
    return this.insert(object);
  }

  return null;
};

// Asynchronous
Brain.prototype.$insertMany = function $insertMany(object) {
  var _this5 = this;

  return new Promise(function (resolve) {
    if (isArray(object)) {
      return resolve(_this5.insert(object));
    }
    return resolve(null);
  });
};

/**
 * find / $find
 * Selects documents in a collection or view and returns a cursor to the selected documents.
 */

// Synchronous
Brain.prototype.find = function find(object) {
  if (isObject(object)) {
    var keys = Object.keys(object);
    var values = Object.values(object);
    var foundById = null;
    var countById = 0;

    if (object && object._id !== undefined) {
      var document = this.store.get(object._id);

      if (document) {
        for (var i = 0; i < keys.length; i += 1) {
          if (document[keys[i]] === values[i]) {
            countById += 1;
          }
        }

        if (countById === keys.length) {
          foundById = document;
        }
      }

      return foundById;
    }

    var found = [];
    this.store.forEach(function (value) {
      var count = 0;

      for (var _i = 0; _i < keys.length; _i += 1) {
        if (value[keys[_i]] === values[_i]) {
          count += 1;
        }
      }

      if (count === keys.length) {
        found.push(value);
      }
    }, this.store);
    return found && found.length > 0 ? found : null;
  } else if (object === undefined) {
    return Array.from(this.store.values());
  }

  return null;
};

// Asynchronous
Brain.prototype.$find = function $find(object) {
  var _this6 = this;

  return new Promise(function (resolve) {
    resolve(_this6.find(object));
  });
};

/**
 * findOne / $findOne
 * Returns one document that satisfies the specified query criteria on the collection or view.
 * If multiple documents satisfy the query, this method returns the first document according to
 * the natural order which reflects the order of documents on the disk. In capped collections,
 * natural order is the same as insertion order. If no document satisfies the query,
 * the method returns null.
 */

// Synchronous
Brain.prototype.findOne = function findOne(object) {
  if (isObject(object)) {
    var keys = Object.keys(object);
    var values = Object.values(object);
    var foundById = null;
    var countById = 0;

    if (object._id) {
      var document = this.store.get(object._id);

      if (document) {
        for (var i = 0; i < keys.length; i += 1) {
          if (document[keys[i]] === values[i]) {
            countById += 1;
          }
        }

        if (countById === keys.length) {
          foundById = document;
        }
      }

      return foundById;
    }

    var found = [];
    this.store.forEach(function (value) {
      var count = 0;

      for (var _i2 = 0; _i2 < keys.length; _i2 += 1) {
        if (value[keys[_i2]] === values[_i2]) {
          count += 1;
        }
      }

      if (count === keys.length && found.length < 1) {
        found.push(value);
      }
    }, this.store);

    return found && found.length > 0 ? found[0] : {};
  } else if (object === undefined) {
    return Array.from(this.store.values())[0] || null;
  }

  return null;
};

// Asynchronous
Brain.prototype.$findOne = function $findOne(object) {
  var _this7 = this;

  return new Promise(function (resolve) {
    resolve(_this7.findOne(object));
  });
};

/**
 * update / $update
 * Modifies an existing document or documents in a collection.
 * The method can modify specific fields of an existing document or
 * documents or replace an existing document entirely, depending on the update parameter.
 */

// Synchronous
Brain.prototype.update = function update(_query, _update, _config) {
  var MergedDocuments = [];
  var documents = this.find(_query);
  if (documents && isArray(documents) && documents.length > 0) {
    for (var i = 0; i < documents.length; i += 1) {
      MergedDocuments.push(Object.assign(documents[i], _update));
    }
    return this.insert(MergedDocuments);
  } else if (documents && isObject(documents)) {
    return this.insert(Object.assign(documents, _update));
  } else if (_config && isObject(_config) && _config.upsert === true) {
    return this.insert(Object.assign(_query, _update));
  }
  return null;
};

// Asynchronous
Brain.prototype.$update = function $update(_query, _update, _config) {
  var _this8 = this;

  return new Promise(function (resolve) {
    _this8.$find(_query).then(function (documents) {
      if (documents && isArray(documents) && documents.length > 0) {
        var MergedDocuments = [];
        for (var i = 0; i < documents.length; i += 1) {
          MergedDocuments.push(Object.assign(documents[i], _update));
        }
        resolve(_this8.$insert(MergedDocuments));
      } else if (documents && isObject(documents)) {
        resolve(_this8.$insert(Object.assign(documents, _update)));
      } else if (_config && isObject(_config) && _config.upsert === true) {
        resolve(_this8.$insert(Object.assign(_query, _update)));
      }
    });
  });
};

/**
 * updateOne / $updateOne
 * Modifies a single existing document in a collection.
 * The method can modify specific fields of an existing document or
 * replace an existing document entirely, depending on the update parameter.
 */

// Synchronous
Brain.prototype.updateOne = function updateOne(_query, _update, _config) {
  var documents = this.find(_query);
  if (documents && isArray(documents) && documents.length > 0) {
    return this.insert(Object.assign(documents[0], _update));
  } else if (documents && isObject(documents)) {
    return this.insert(Object.assign(documents, _update));
  } else if (_config && isObject(_config) && _config.upsert === true) {
    return this.insert(Object.assign(_query, _update));
  }
  return null;
};

// Asynchronous
Brain.prototype.$updateOne = function $updateOne(_query, _update, _config) {
  var _this9 = this;

  return new Promise(function (resolve) {
    _this9.$find(_query).then(function (documents) {
      if (documents && isArray(documents) && documents.length > 0) {
        return resolve(_this9.$insert(Object.assign(documents[0], _update)));
      } else if (documents && isObject(documents)) {
        return resolve(_this9.$insert(Object.assign(documents, _update)));
      } else if (_config && isObject(_config) && _config.upsert === true) {
        return resolve(_this9.$insert(Object.assign(_query, _update)));
      }
      return resolve(null);
    });
  });
};

/**
 * updateMany / $updateMany
 * Modifies an existing document or documents in a collection.
 * The method can modify specific fields of an existing document or
 * documents or replace an existing document entirely, depending on the update parameter.
 */

// Synchronous
Brain.prototype.updateMany = Brain.prototype.update;

// Asynchronous
Brain.prototype.$updateMany = Brain.prototype.$update;

// Synchronous
Brain.prototype.count = function count() {
  return this.store.size;
};

// Asynchronous
Brain.prototype.$count = function $count() {
  var _this10 = this;

  return new Promise(function (resolve) {
    resolve(_this10.store.size);
  });
};

// Synchronous
Brain.prototype.stats = function stats() {
  var store = this.store;
  return readDatabaseStatsSync(store, this.filePath, this.fileName);
};

// Asynchronous
Brain.prototype.$stats = function $stats() {
  var _this11 = this;

  var store = this.store;
  return new Promise(function (resolve) {
    resolve(readDatabaseStats(store, _this11.filePath, _this11.fileName));
  });
};

/*

// TODO
Brain.prototype.renameCollection = function renameCollection(index, value) {
  return this.store.set(index, value);
};

Brain.prototype.drop = function drop() {
  this.store.clear();
  return this.store;
};

*/

/**
 * deleteOne deletes the first document that matches the filter.
 * Use a field that is part of a unique index such as _id for precise deletions.
 */

/*
Brain.prototype.deleteOne = function deleteOne(object) {
  if (object) {
    const keys = Object.keys(object);
    const values = Object.values(object);
    let found = 0;
    let index = null;

    this.store.forEach((value, key) => {
      let count = 0;

      for (let i = 0; i < keys.length; i += 1) {
        if (value[keys[i]] === values[i]) {
          count += 1;
        }
      }

      if (count === keys.length && found < 1) {
        found += 1;
        index = key;
      }
    }, this.store);

    if (index !== null) {
      return this.store.delete(index);
    }
    return false;
  }
  return false;
};
*/

/**
 * Removes all documents that match the filter from a collection.
 */

/*
Brain.prototype.deleteMany = function deleteMany(object) {
  if (object) {
    const keys = Object.keys(object);
    const values = Object.values(object);
    const found = [];

    this.store.forEach((value, key) => {
      let count = 0;

      for (let i = 0; i < keys.length; i += 1) {
        if (value[keys[i]] === values[i]) {
          count += 1;
        }
      }

      if (count === keys.length) {
        found.push(key);
      }
    }, this.store);

    if (found.length > 0) {
      for (let i = 0; i < found.length; i += 1) {
        this.store.delete(found[i]);
      }
      return true;
    }
    return false;
  }
  return false;
};

Brain.prototype.deleteOneById = function deleteOne(index) {
  return this.store.delete(index);
};
// TODO
Brain.prototype.findAndModify = function findAndModify(object) {
  return object;
};
// TODO
Brain.prototype.findOneAndDelete = function findOneAndDelete(object) {
  return object;
};
// TODO
Brain.prototype.findOneAndReplace = function findOneAndReplace(object) {
  return object;
};
// TODO
Brain.prototype.findOneAndUpdate = function findOneAndUpdate(object) {
  return object;
};

Brain.prototype.getIndexes = function getIndexes() {
  return Array.from(this.store.keys());
};
*/

var main = Brain;

exports.default = main;

module.exports = main; // for CommonJS compatibility